@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 1993-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 1993-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -48,6 +48,8 @@
 
 static void RmUnmapBusAperture (OBJGPU *, NvP64, NvU64, NvBool, NvP64);
 
+#include "gpu/conf_compute/conf_compute.h"
+
 typedef struct RS_CPU_MAP_PARAMS RmMapParams;
 typedef struct RS_CPU_UNMAP_PARAMS RmUnmapParams;
 
@@ -152,6 +154,14 @@ rmapiGetEffectiveAddrSpace
     return NV_OK;
 }
 
+// Asserts to check caching type matches across sdk and nv_memory_types
+ct_assert(NVOS33_FLAGS_CACHING_TYPE_CACHED        == NV_MEMORY_CACHED);
+ct_assert(NVOS33_FLAGS_CACHING_TYPE_UNCACHED      == NV_MEMORY_UNCACHED);
+ct_assert(NVOS33_FLAGS_CACHING_TYPE_WRITECOMBINED == NV_MEMORY_WRITECOMBINED);
+ct_assert(NVOS33_FLAGS_CACHING_TYPE_WRITEBACK     == NV_MEMORY_WRITEBACK);
+ct_assert(NVOS33_FLAGS_CACHING_TYPE_DEFAULT       == NV_MEMORY_DEFAULT);
+ct_assert(NVOS33_FLAGS_CACHING_TYPE_UNCACHED_WEAK == NV_MEMORY_UNCACHED_WEAK);
+
 //
 // Map memory entry points.
 //
@@ -168,8 +178,7 @@ memMap_IMPL
     KernelBus *pKernelBus = NULL;
     MemoryManager *pMemoryManager = NULL;
     KernelMemorySystem *pKernelMemorySystem = NULL;
-    RsClient *pRsClient;
-    RmClient *pRmClient;
+    RmClient *pClient;
     RsResourceRef *pContextRef;
     RsResourceRef *pMemoryRef;
     Memory *pMemoryInfo; // TODO: rename this field. pMemoryInfo is the legacy name.
@@ -181,6 +190,8 @@ memMap_IMPL
     NvBool bBroadcast;
     NvU64 mapLimit;
     NvBool bIsSysmem = NV_FALSE;
+    NvBool bSkipSizeCheck = (DRF_VAL(OS33, _FLAGS, _SKIP_SIZE_CHECK, pMapParams->flags) ==
+                             NVOS33_FLAGS_SKIP_SIZE_CHECK_ENABLE);
 
     NV_ASSERT_OR_RETURN(RMCFG_FEATURE_KERNEL_RM, NV_ERR_NOT_SUPPORTED);
 
@@ -196,8 +207,10 @@ memMap_IMPL
         pKernelMemorySystem = GPU_GET_KERNEL_MEMORY_SYSTEM(pGpu);
     }
 
-    NV_ASSERT_OK_OR_RETURN(serverGetClientUnderLock(&g_resServ, pMapParams->hClient, &pRsClient));
-    NV_ASSERT_OK_OR_RETURN(clientGetResourceRef(pRsClient, pMapParams->hMemory, &pMemoryRef));
+    pClient = serverutilGetClientUnderLock(pMapParams->hClient);
+    NV_ASSERT_OR_ELSE(pClient != NULL, return NV_ERR_INVALID_CLIENT);
+    NV_ASSERT_OK_OR_RETURN(clientGetResourceRef(staticCast(pClient, RsClient),
+                pMapParams->hMemory, &pMemoryRef));
 
     pMemoryInfo = dynamicCast(pMemoryRef->pResource, Memory);
     NV_ASSERT_OR_RETURN(pMemoryInfo != NULL, NV_ERR_NOT_SUPPORTED);
@@ -218,11 +231,31 @@ memMap_IMPL
     //
     // CPU to directly access protected memory is allowed on MODS
     //
-    if ((pMemoryInfo->Flags & NVOS32_ALLOC_FLAGS_PROTECTED) &&
-        (pMapParams->protect != NV_PROTECT_WRITEABLE) &&
-        ! RMCFG_FEATURE_PLATFORM_MODS)
+    // The check below is for VPR and should be skipped for Hopper CC
+    if ((pGpu != NULL) && !gpuIsCCFeatureEnabled(pGpu))
     {
-        return NV_ERR_NOT_SUPPORTED;
+        if ((pMemoryInfo->Flags & NVOS32_ALLOC_FLAGS_PROTECTED) &&
+            (pMapParams->protect != NV_PROTECT_WRITEABLE) &&
+            ! RMCFG_FEATURE_PLATFORM_MODS)
+        {
+            return NV_ERR_NOT_SUPPORTED;
+        }
+    }
+
+    if ((pGpu != NULL) && (pMemoryInfo->Flags & NVOS32_ALLOC_FLAGS_PROTECTED))
+    {
+        ConfidentialCompute *pCC = GPU_GET_CONF_COMPUTE(pGpu);
+        //
+        // If neither BAR1 nor PCIE as a whole is trusted, fail the mapping
+        // for allocations in CPR region. Mapping should still succeed for
+        // allocations in non-CPR region
+        //
+        if ((pCC != NULL) && !pCC->ccStaticInfo.bIsBar1Trusted &&
+            !pCC->ccStaticInfo.bIsPcieTrusted)
+        {
+            NV_ASSERT(0);
+            return NV_ERR_NOT_SUPPORTED;
+        }
     }
 
     if (!pMapParams->bKernel &&
@@ -242,14 +275,18 @@ memMap_IMPL
         return NV_ERR_INVALID_LIMIT;
     }
 
+    if (bSkipSizeCheck && (pCallContext->secInfo.privLevel < RS_PRIV_LEVEL_KERNEL))
+    {
+        return NV_ERR_INSUFFICIENT_PERMISSIONS;
+    }
+
     //
     // See bug #140807 and #150889 - we need to pad memory mappings to past their
     // actual allocation size (to PAGE_SIZE+1) because of a buggy ms function so
     // skip the allocation size sanity check so the map operation still succeeds.
     //
-    if ((DRF_VAL(OS33, _FLAGS, _SKIP_SIZE_CHECK, pMapParams->flags) == NVOS33_FLAGS_SKIP_SIZE_CHECK_DISABLE) &&
-        (!portSafeAddU64(pMapParams->offset, pMapParams->length, &mapLimit) ||
-         (mapLimit > pMemoryInfo->Length)))
+    if (!portSafeAddU64(pMapParams->offset, pMapParams->length, &mapLimit) ||
+        (!bSkipSizeCheck && (mapLimit > pMemoryInfo->Length)))
     {
         return NV_ERR_INVALID_LIMIT;
     }
@@ -264,6 +301,7 @@ memMap_IMPL
     }
 
     bIsSysmem = (effectiveAddrSpace == ADDR_SYSMEM);
+    bIsSysmem = bIsSysmem || (effectiveAddrSpace == ADDR_EGM);
 
     if (dynamicCast(pMemoryInfo, FlaMemory) != NULL)
     {
@@ -317,7 +355,6 @@ memMap_IMPL
             }
             else
             {
-                NV_ASSERT(DRF_VAL(OS33, _FLAGS, _BUS, pMapParams->flags) == NVOS33_FLAGS_BUS_NVLINK_COHERENT);
 
                 //
                 // Allocating mapping for user mode client
@@ -423,12 +460,15 @@ memMap_IMPL
             NV_ASSERT(pGpu->busInfo.gpuPhysFbAddr);
 
             {
+                Device *pDevice = NULL;
+
                 // Below, we only map one GPU's address for CPU access, so we can use UNICAST here
                 NvU32 busMapFbFlags = BUS_MAP_FB_FLAGS_MAP_UNICAST;
                 if(DRF_VAL(OS33, _FLAGS, _MAPPING, pMapParams->flags) == NVOS33_FLAGS_MAPPING_DIRECT)
                 {
                     busMapFbFlags |= BUS_MAP_FB_FLAGS_DISABLE_ENCRYPTION;
                 }
+
                 switch (pMapParams->protect)
                 {
                     case NV_PROTECT_READABLE:
@@ -441,10 +481,20 @@ memMap_IMPL
 
                 pMemDesc = memdescGetMemDescFromGpu(pMemDesc, pGpu);
 
+                // WAR for Bug 3564398, need to allocate doorbell for windows differently
+                if (RMCFG_FEATURE_PLATFORM_WINDOWS_LDDM &&
+                    memdescGetFlag(pMemDesc, MEMDESC_FLAGS_MAP_SYSCOH_OVER_BAR1))
+                {
+                    busMapFbFlags |= BUS_MAP_FB_FLAGS_MAP_DOWNWARDS;
+                }
+
+                (void) deviceGetByHandle(staticCast(pClient, RsClient),
+                                         pMapParams->hDevice, &pDevice);
+
                 rmStatus = kbusMapFbAperture_HAL(pGpu, pKernelBus,
                                                  pMemDesc, pMapParams->offset,
                                                  &gpuVirtAddr, &gpuMapLength,
-                                                 busMapFbFlags, pMapParams->hClient);
+                                                 busMapFbFlags, pDevice);
             }
 
             if (rmStatus != NV_OK)
@@ -487,6 +537,7 @@ memMap_IMPL
         // direct mapping.
         //
         pMapParams->flags = FLD_SET_DRF(OS33, _FLAGS, _MAPPING, _REFLECTED, pMapParams->flags);
+        pMapParams->flags = FLD_SET_DRF_NUM(OS33, _FLAGS, _CACHING_TYPE, cachingType, pMapParams->flags);
 
         if (rmStatus != NV_OK)
             goto _rmMapMemory_pciFail;
@@ -561,13 +612,12 @@ memMap_IMPL
     {
         RS_PRIV_LEVEL privLevel;
 
-        pRmClient = dynamicCast(pRsClient, RmClient);
-        if (pRmClient == NULL)
-            return NV_ERR_OPERATING_SYSTEM;
-
-        privLevel = rmclientGetCachedPrivilege(pRmClient);
-        if (!rmclientIsAdmin(pRmClient, privLevel) && !memdescGetFlag(pMemDesc, MEMDESC_FLAGS_SKIP_REGMEM_PRIV_CHECK))
+        privLevel = rmclientGetCachedPrivilege(pClient);
+        if (!rmclientIsAdmin(pClient, privLevel) &&
+            !memdescGetFlag(pMemDesc, MEMDESC_FLAGS_SKIP_REGMEM_PRIV_CHECK))
+        {
             return NV_ERR_PROTECTION_FAULT;
+        }
 
         if (DRF_VAL(OS33, _FLAGS, _MEM_SPACE, pMapParams->flags) == NVOS33_FLAGS_MEM_SPACE_USER)
         {
@@ -683,8 +733,9 @@ memUnmap_IMPL
         //
     }
     // System Memory case
-    else if ((pGpu == NULL) || ((memdescGetAddressSpace(pMemDesc) == ADDR_SYSMEM) &&
-                                 FLD_TEST_DRF(OS33, _FLAGS, _MAPPING, _DIRECT, pCpuMapping->flags)))
+    else if ((pGpu == NULL) || (((memdescGetAddressSpace(pMemDesc) == ADDR_SYSMEM)
+                                 || (memdescGetAddressSpace(pMemDesc) == ADDR_EGM)
+                                ) && FLD_TEST_DRF(OS33, _FLAGS, _MAPPING, _DIRECT, pCpuMapping->flags)))
     {
         if (FLD_TEST_DRF(OS33, _FLAGS, _MAPPING, _DIRECT, pCpuMapping->flags))
         {
@@ -793,8 +844,7 @@ serverMap_Prologue
 )
 {
     NV_STATUS           rmStatus;
-    RsClient           *pRsClient;
-    RmClient           *pRmClient;
+    RmClient           *pClient;
     RsResourceRef      *pMemoryRef;
     NvHandle            hClient = pMapParams->hClient;
     NvHandle            hParent = hClient;
@@ -808,16 +858,15 @@ serverMap_Prologue
         return NV_ERR_INVALID_FLAGS;
 
     // Populate Resource Server information
-    NV_ASSERT_OK_OR_RETURN(serverGetClientUnderLock(&g_resServ, hClient, &pRsClient));
+    pClient = serverutilGetClientUnderLock(hClient);
+    NV_ASSERT_OR_ELSE(pClient != NULL, return NV_ERR_INVALID_CLIENT);
 
     // Validate hClient
-    pRmClient = dynamicCast(pRsClient, RmClient);
-    if (pRmClient == NULL)
-        return NV_ERR_OPERATING_SYSTEM;
-    privLevel = rmclientGetCachedPrivilege(pRmClient);
+    privLevel = rmclientGetCachedPrivilege(pClient);
 
     // RS-TODO: Assert if this fails after all objects are converted
-    NV_ASSERT_OK_OR_RETURN(clientGetResourceRef(pRsClient, pMapParams->hMemory, &pMemoryRef));
+    NV_ASSERT_OK_OR_RETURN(clientGetResourceRef(staticCast(pClient, RsClient),
+                pMapParams->hMemory, &pMemoryRef));
 
     if (pMemoryRef->pParentRef != NULL)
         hParent = pMemoryRef->pParentRef->hResource;
@@ -837,7 +886,9 @@ serverMap_Prologue
         NV_ASSERT_OR_RETURN(hParent != hClient, NV_ERR_INVALID_OBJECT_PARENT);
 
         RsResourceRef *pContextRef;
-        rmStatus = clientGetResourceRef(pRsClient, pMapParams->hDevice, &pContextRef);
+        rmStatus = clientGetResourceRef(staticCast(pClient, RsClient),
+                pMapParams->hDevice, &pContextRef);
+
         if (rmStatus != NV_OK)
             return rmStatus;
 
@@ -895,8 +946,7 @@ serverUnmap_Prologue
 {
     OBJGPU *pGpu = NULL;
     NV_STATUS rmStatus;
-    RsClient *pRsClient;
-    RmClient *pRmClient;
+    RmClient *pClient;
     RsResourceRef *pMemoryRef;
     NvHandle hClient = pUnmapParams->hClient;
     NvHandle hParent = hClient;
@@ -909,16 +959,15 @@ serverUnmap_Prologue
     void *pProcessHandle = NULL;
 
     // Populate Resource Server information
-    NV_ASSERT_OK_OR_RETURN(serverGetClientUnderLock(&g_resServ, hClient, &pRsClient));
+    pClient = serverutilGetClientUnderLock(hClient);
+    NV_ASSERT_OR_ELSE(pClient != NULL, return NV_ERR_INVALID_CLIENT);
 
     // check if we have a user or kernel RM client
-    pRmClient = dynamicCast(pRsClient, RmClient);
-    if (pRmClient == NULL)
-        return NV_ERR_OPERATING_SYSTEM;
-    privLevel = rmclientGetCachedPrivilege(pRmClient);
+    privLevel = rmclientGetCachedPrivilege(pClient);
 
     // RS-TODO: Assert if this fails after all objects are converted
-    NV_ASSERT_OK_OR_RETURN(clientGetResourceRef(pRsClient, hMemory, &pMemoryRef));
+    NV_ASSERT_OK_OR_RETURN(clientGetResourceRef(staticCast(pClient, RsClient),
+                hMemory, &pMemoryRef));
 
     if (pMemoryRef->pParentRef != NULL)
         hParent = pMemoryRef->pParentRef->hResource;
@@ -933,7 +982,9 @@ serverUnmap_Prologue
         NV_ASSERT_OR_RETURN(hParent != hClient, NV_ERR_INVALID_OBJECT_PARENT);
 
         RsResourceRef *pContextRef;
-        rmStatus = clientGetResourceRef(pRsClient, pUnmapParams->hDevice, &pContextRef);
+        rmStatus = clientGetResourceRef(staticCast(pClient, RsClient),
+                pUnmapParams->hDevice, &pContextRef);
+
         if (rmStatus != NV_OK)
             return rmStatus;
 
@@ -974,11 +1025,24 @@ serverUnmap_Prologue
     {
         rmStatus = osAttachToProcess(&pProcessHandle, ProcessId);
         if (rmStatus != NV_OK)
-            return rmStatus;
+        {
+            if (pUnmapParams->bTeardown)
+                pProcessHandle = NULL;
+            else
+                return rmStatus;
+        }
 
         pUnmapParams->pProcessHandle = pProcessHandle;
     }
 
+    // Don't do any filtering if this is a tear-down path
+    if (pUnmapParams->bTeardown)
+    {
+        pUnmapParams->fnFilter = NULL;
+        return NV_OK;
+    }
+
+
     pUnmapParams->fnFilter = bKernel
         ? serverutilMappingFilterKernel
         : serverutilMappingFilterCurrentUserProc;
@@ -1048,7 +1112,6 @@ rmapiMapToCpu
     return status;
 }
 
-
 /**
  * Call into Resource Server to register and execute a CPU mapping operation.
  *
@@ -1063,7 +1126,7 @@ rmapiMapToCpu
  *    6. Release any locks taken
  */
 NV_STATUS
-rmapiMapToCpuWithSecInfo
+rmapiMapToCpuWithSecInfoV2
 (
     RM_API            *pRmApi,
     NvHandle           hClient,
@@ -1072,7 +1135,7 @@ rmapiMapToCpuWithSecInfo
     NvU64              offset,
     NvU64              length,
     NvP64             *ppCpuVirtAddr,
-    NvU32              flags,
+    NvU32             *flags,
     API_SECURITY_INFO *pSecInfo
 )
 {
@@ -1086,13 +1149,13 @@ rmapiMapToCpuWithSecInfo
               hDevice, hMemory);
     NV_PRINTF(LEVEL_INFO,
               "Nv04MapMemory:  offset: %llx length: %llx flags:0x%x\n",
-              offset, length, flags);
+              offset, length, *flags);
 
     status = rmapiPrologue(pRmApi, &rmApiContext);
     if (status != NV_OK)
         return status;
 
-    NV_PRINTF(LEVEL_INFO, "MMU_PROFILER Nv04MapMemory 0x%x\n", flags);
+    NV_PRINTF(LEVEL_INFO, "MMU_PROFILER Nv04MapMemory 0x%x\n", *flags);
 
     portMemSet(&lockInfo, 0, sizeof(lockInfo));
     rmapiInitLockInfo(pRmApi, hClient, &lockInfo);
@@ -1109,7 +1172,7 @@ rmapiMapToCpuWithSecInfo
     rmMapParams.offset = offset;
     rmMapParams.length = length;
     rmMapParams.ppCpuVirtAddr = ppCpuVirtAddr;
-    rmMapParams.flags = flags;
+    rmMapParams.flags = *flags;
     rmMapParams.pLockInfo = &lockInfo;
     rmMapParams.pSecInfo = pSecInfo;
 
@@ -1117,6 +1180,8 @@ rmapiMapToCpuWithSecInfo
 
     rmapiEpilogue(pRmApi, &rmApiContext);
 
+    *flags = rmMapParams.flags;
+
     if (status == NV_OK)
     {
         NV_PRINTF(LEVEL_INFO, "Nv04MapMemory: complete\n");
@@ -1134,6 +1199,25 @@ rmapiMapToCpuWithSecInfo
     return status;
 }
 
+NV_STATUS
+rmapiMapToCpuWithSecInfo
+(
+    RM_API            *pRmApi,
+    NvHandle           hClient,
+    NvHandle           hDevice,
+    NvHandle           hMemory,
+    NvU64              offset,
+    NvU64              length,
+    NvP64             *ppCpuVirtAddr,
+    NvU32              flags,
+    API_SECURITY_INFO *pSecInfo
+)
+{
+    return rmapiMapToCpuWithSecInfoV2(pRmApi, hClient,
+        hDevice, hMemory, offset, length, ppCpuVirtAddr,
+        &flags, pSecInfo);
+}
+
 NV_STATUS
 rmapiMapToCpuWithSecInfoTls
 (
@@ -1153,7 +1237,32 @@ rmapiMapToCpuWithSecInfoTls
 
     threadStateInit(&threadState, THREAD_STATE_FLAGS_NONE);
 
-    status = rmapiMapToCpuWithSecInfo(pRmApi, hClient, hDevice, hMemory, offset, length, ppCpuVirtAddr, flags, pSecInfo);
+    status = rmapiMapToCpuWithSecInfoV2(pRmApi, hClient, hDevice, hMemory, offset, length, ppCpuVirtAddr, &flags, pSecInfo);
+
+    threadStateFree(&threadState, THREAD_STATE_FLAGS_NONE);
+
+    return status;
+}
+NV_STATUS
+rmapiMapToCpuWithSecInfoTlsV2
+(
+    RM_API            *pRmApi,
+    NvHandle           hClient,
+    NvHandle           hDevice,
+    NvHandle           hMemory,
+    NvU64              offset,
+    NvU64              length,
+    NvP64             *ppCpuVirtAddr,
+    NvU32             *flags,
+    API_SECURITY_INFO *pSecInfo
+)
+{
+    THREAD_STATE_NODE threadState;
+    NV_STATUS         status;
+
+    threadStateInit(&threadState, THREAD_STATE_FLAGS_NONE);
+
+    status = rmapiMapToCpuWithSecInfoV2(pRmApi, hClient, hDevice, hMemory, offset, length, ppCpuVirtAddr, flags, pSecInfo);
 
     threadStateFree(&threadState, THREAD_STATE_FLAGS_NONE);
 
